#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
import pwd
import grp
import os
import logging
from cfnbootstrap.construction_errors import ToolError
from cfnbootstrap.util import ProcessHelper

log = logging.getLogger("cfn.init")

def set_owner_and_group(filename, owner_name, group_name):
    owner_id = -1
    group_id = -1

    if owner_name:
        try:
            owner_id = pwd.getpwnam(owner_name)[2]
        except KeyError:
            raise ToolError("%s is not a valid user name" % owner_name)

    if group_name:
        try:
            group_id = grp.getgrnam(group_name)[2]
        except KeyError:
            raise ToolError("%s is not a valid group name" % group_name)

    if group_id != -1 or owner_id != -1:
        logging.debug("Setting owner %s and group %s for %s", owner_id, group_id, filename)
        os.lchown(filename, owner_id, group_id)

def create_group(group_name, gid=None):
    """Create a group in the OS, returning True if one is created"""
    try:
        group_record = grp.getgrnam(group_name)
        if gid and str(group_record[2]) != gid:
            raise ToolError("Group %s exists with gid %s, but gid %s was requested" % (group_name, group_record[2], gid))
        log.debug("Group %s already exists", group_name)
        return False
    except KeyError:
        pass

    cmd = ['/usr/sbin/groupadd', '-r']

    if gid:
        cmd.extend(['-g', gid])

    cmd.append(group_name)

    result = ProcessHelper(cmd).call()

    if result.returncode:
        log.error("Failed to create group %s", group_name)
        log.debug("Groupadd output: %s", result.stdout)
        raise ToolError("Failed to create group %s" % group_name)
    else:
        log.info("Created group %s successfully", group_name)
        return True


def create_or_modify_user(user_name, groups=[], homedir=None, uid=None):
    """Create or modify a user in the OS, returning True if action was taken"""
    try:
        user_record = pwd.getpwnam(user_name)
        if uid and str(user_record[2]) != uid:
            raise ToolError("User %s exists with uid %s, but uid %s was requested" % (user_name, user_record[2], uid))
        return _modify_user(user_name, groups, homedir)
    except KeyError:
        _create_user(user_name, groups, homedir, uid)
        return True

def _modify_user(user_name, groups=[], homedir=None):
    """ Modify a user and return True, else return False """
    if not homedir and not groups:
        log.info("No homedir or groups specified; not modifying %s", user_name)
        return False

    cmd = ['/usr/sbin/usermod']

    if groups:
        gids = _get_gids(groups)
        current_gids = _gids_for_user(user_name)
        if frozenset(gids) ^ frozenset(current_gids):
            cmd.extend(['-G', ','.join(gids)])
        else:
            log.debug("Groups have not changed for %s", user_name)

    if homedir:
        if homedir != _get_user_homedir(user_name):
            cmd.extend(['-d', homedir])
        else:
            log.debug("Homedir has not changed for %s", user_name)

    if len(cmd) == 1:
        log.debug("User %s does not need modification", user_name)
        return False

    cmd.append(user_name)

    result = ProcessHelper(cmd).call()

    if result.returncode:
        log.error("Failed to modify user %s", user_name)
        log.debug("Usermod output: %s", result.stdout)
        raise ToolError("Failed to modify user %s" % user_name)
    else:
        log.info("Modified user %s successfully", user_name)
        return True

def _get_user_homedir(user_name):
    return pwd.getpwnam(user_name)[5]

def _create_user(user_name, groups=[], homedir=None, uid=None):
    gids = _get_gids(groups)

    cmd = ['/usr/sbin/useradd', '-M', '-r', '--shell', '/sbin/nologin']

    if homedir:
        cmd.extend(['-d', homedir])

    if uid:
        cmd.extend(['-u', uid])

    if gids:
        cmd.extend(['-G', ','.join(gids)])

    cmd.append(user_name)

    result = ProcessHelper(cmd).call()

    if result.returncode:
        log.error("Failed to add user %s", user_name)
        log.debug("Useradd output: %s", result.stdout)
        raise ToolError("Failed to add user %s" % user_name)
    else:
        log.info("Added user %s successfully", user_name)

def _get_gids(groups):
    gids = []
    for group_name in groups:
        try:
            gids.append(str(grp.getgrnam(group_name)[2]))
        except KeyError:
            raise ToolError("%s is not a valid group name" % group_name)
    return gids

def _gids_for_user(user_name):
    return [str(group[2]) for group in grp.getgrall() if user_name in group[3]]



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
"""
Base classes for AWS/QUERY clients

Classes:
AWSClient - an HTTP client that makes signed requests

"""
from cfnbootstrap import util
from cfnbootstrap.packages.requests import api
from xml.etree import ElementTree
import StringIO
import base64
import datetime
import hashlib
import hmac
import logging
import operator
import re
import urllib
import urlparse
import public_constants
import sys
import platform

log = logging.getLogger('cfn.client')

class Signer(object):

    def sign(self, verb, base_url, params, creds, in_headers=None, timestamp=None):
        pass

    def _normalize_url(self, base_url):
        return base_url if base_url.endswith('/') else base_url + '/'

class CFNSigner(Signer):

    def sign(self, verb, base_url, params, creds, in_headers=None, timestamp=None):
        base_url = self._normalize_url(base_url)

        if not util.is_ec2():
            raise ValueError("Cannot use CFN signature outside of EC2")

        document = util.get_instance_identity_document()
        signature = util.get_instance_identity_signature()

        new_headers = dict({} if in_headers is None else in_headers)
        new_headers['Authorization'] = 'CFN_V1 %s:%s' % (base64.b64encode(document), signature.replace('\n', ''))

        return (verb, base_url, params, new_headers)

class V4Signer(Signer):

    def __init__(self, region, service, terminator='aws4_request'):
        self._region = region
        self._service = service
        self._terminator = terminator

    def sign(self, verb, base_url, params, creds, in_headers=None, timestamp=None):
        if not in_headers: in_headers = {}
        base_url = self._normalize_url(base_url)

        if not timestamp:
            timestamp = datetime.datetime.utcnow()

        new_headers = dict(in_headers)

        timestamp_formatted = timestamp.strftime('%Y%m%dT%H%M%SZ')
        timestamp_short = timestamp.strftime('%Y%m%d')

        scope =  timestamp_short + '/' + self._region + '/' + self._service + '/' + self._terminator

        if 'Date' in new_headers:
            del new_headers['Date']
        new_headers['X-Amz-Date'] = timestamp_formatted
        if creds.security_token:
            new_headers['X-Amz-Security-Token'] = creds.security_token
        new_headers['Host'] = urlparse.urlsplit(base_url).netloc
        if verb == 'POST':
            new_headers['Content-type'] = 'application/x-www-form-urlencoded'

        canonical_request = verb + '\n'
        canonical_request += self._canonicalize_uri(base_url) + '\n'
        canonical_request += (self._canonicalize_query(params) if verb == 'GET' else '') + '\n'

        (canonical_headers, signed_headers) = self._canonicalize_headers(new_headers)
        canonical_request += canonical_headers + '\n' + signed_headers + '\n'
        canonical_request += hashlib.sha256(Client.construct_query(params).encode('utf-8') if verb == 'POST' else '').hexdigest()

        string_to_sign = 'AWS4-HMAC-SHA256\n' + timestamp_formatted + '\n' + scope + '\n' + hashlib.sha256(canonical_request.encode('utf-8')).hexdigest()

        derived_key = hmac.new(("AWS4" + creds.secret_key).encode('utf-8'), timestamp_short.encode('utf-8'), hashlib.sha256).digest()
        derived_key = hmac.new(derived_key, self._region.encode('utf-8'), hashlib.sha256).digest()
        derived_key = hmac.new(derived_key, self._service.encode('utf-8'), hashlib.sha256).digest()
        derived_key = hmac.new(derived_key, "aws4_request".encode('utf-8'), hashlib.sha256).digest()

        signature = hmac.new(derived_key, string_to_sign.encode('utf-8'), hashlib.sha256).hexdigest()

        credential = creds.access_key + '/' + scope
        new_headers['Authorization'] = 'AWS4-HMAC-SHA256 Credential=%s,SignedHeaders=%s, Signature=%s' % (credential, signed_headers, signature)

        return (verb, base_url, params, new_headers)

    def _canonicalize_uri(self, uri):
        split = urlparse.urlsplit(uri)
        if not split.path:
            return '/'
        path = urlparse.urlsplit(urlparse.urljoin('http://foo.com', split.path.lstrip('/'))).path
        return urllib.quote(path, '/~') if path else '/'

    def _canonicalize_query(self, params):
        if not params:
            return ''

        encoded_pairs = ((urllib.quote(entry[0], '~'), urllib.quote(entry[1], '~') if len(entry) > 1 else '') for entry in params.iteritems())
        sorted_pairs = sorted(encoded_pairs, key=operator.itemgetter(0, 1))

        return '&'.join(('='.join(pair) for pair in sorted_pairs))

    def _canonicalize_headers(self, headers):
        canon_headers = {}
        for key, value in ((key.lower(), re.sub(r'(?su)[\s]+', ' ', value).strip()) for key, value in headers.iteritems()):
            if key in canon_headers:
                canon_headers[key] = canon_headers[key] + ',' + value
            else:
                canon_headers[key] = value

        sorted_entries = sorted(canon_headers.iteritems(), key=operator.itemgetter(0))

        return '\n'.join((':'.join(entry) for entry in sorted_entries)) + '\n', ';'.join((entry[0] for entry in sorted_entries))


class AwsQueryError(util.RemoteError):

    def __init__(self, status_code, error_code, error_type, msg):
        # Retry for Throttling or InvalidAccessKeyId (IAM propagation delay)
        if status_code == 503 or error_code in ('Throttling', 'InvalidAccessKeyId', 'InvalidClientTokenId'):
            retry_mode = 'RETRIABLE_FOREVER'
        elif error_type == 'Sender':
            retry_mode = 'TERMINAL'
        else:
            retry_mode = 'RETRIABLE'

        super(AwsQueryError, self).__init__(status_code, "%s: %s" % (error_code, msg), retry_mode)

        self.error_code = error_code
        self.error_type = error_type

class Client(object):
    '''
    A base AWS/QUERY client
    '''

    def __init__(self, credentials, is_json, endpoint=None, signer=None, xmlns=None, proxyinfo=None):
        self._credentials = credentials
        self._endpoint = endpoint
        self._is_json = is_json
        self._xmlns = xmlns
        if not signer:
            raise ValueError('A valid signer is required')
        self._signer = signer
        self._proxyinfo = dict(proxyinfo) if proxyinfo else None


    @staticmethod
    def construct_query(sign_data):
        ret_str = ''
        for k, vs in sorted(sign_data.iteritems(), key=operator.itemgetter(0)):
            if isinstance(vs, list):
                for v in sorted(vs):
                    ret_str += '&'.join(urllib.quote(k, safe='~') + '=' + urllib.quote(v, safe='~'))
            else:
                if ret_str:
                    ret_str += '&'
                ret_str += urllib.quote(k, safe='~') + '=' + urllib.quote(vs, safe='~')

        return ret_str

    @staticmethod
    def _extract_json_message(resp):
        try:
            eDoc = resp.json()['Error']
            code = eDoc['Code']
            message = eDoc['Message']
            error_type = eDoc['Type']

            return AwsQueryError(resp.status_code, code, error_type, message)
        except (TypeError, AttributeError, KeyError, ValueError):
            return AwsQueryError(resp.status_code, 'Unknown', 'Receiver', resp.text)

    @staticmethod
    def _get_xml_extractor(xmlns):
        def _extract_xml_message(resp):
            try:
                eDoc = ElementTree.ElementTree(file=StringIO.StringIO(resp.content))
                code = eDoc.findtext('{%s}Error/{%s}Code' % (xmlns, xmlns))
                error_type = eDoc.findtext('{%s}Error/{%s}Type' % (xmlns, xmlns))
                message = eDoc.findtext('{%s}Error/{%s}Message' % (xmlns, xmlns))

                return AwsQueryError(resp.status_code, code, error_type, message)
            except (TypeError, AttributeError, KeyError, ValueError):
                return AwsQueryError(resp.status_code, 'Unknown', 'Receiver', resp.text)

        return _extract_xml_message

    def _call(self, params, endpoint=None, request_credentials=None, verb='GET', timeout=None):
        base = endpoint if endpoint else self._endpoint
        creds = request_credentials if request_credentials else self._credentials
        accept_type = "application/json" if self._is_json else "application/xml"
        req = self._signer.sign(verb, base, params, creds, {"Accept" : accept_type})

        return self._make_request(*req, timeout=timeout)

    def _make_request(self, verb, base_url, params, headers, timeout=None):
        headers = dict(headers) if headers else {}
        headers['User-Agent'] = self._construct_user_agent()

        return util.check_status(api.request(verb, base_url,
                           **util.req_opts({
                            'data' : Client.construct_query(params) if verb=='POST' else dict(),
                            'params' : params if verb!='POST' else dict(),
                            'headers' : headers,
                            'proxies' : self._proxyinfo,
                            'timeout' : timeout
                           })))

    def _construct_user_agent(self):
        # User-agent format: "CfnTools/[version] (system information) python/[version]"
        # In case it fails to get any part of the string above,
        # it defaults back to the old string - "CloudFormation Tools"
        try:
            bootstrapper_version = public_constants.get_version()
            python_version = platform.python_version()
            system_info = platform.platform()
            return 'CfnTools/' + str(bootstrapper_version) + ' (' + system_info + ')' +' python/' + python_version
        except Exception as e:
            log.debug("Constructing user agent failed with error message: " + str(e))
            return "CloudFormation Tools"



#!/usr/bin/env python

#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================

from distutils.core import setup, Distribution
import sys

name = 'aws-cfn-bootstrap'
version = '1.4-31'

if sys.version_info[0] == 2 and sys.version_info[1] < 6:
        print >> sys.stderr, "Python 2.6+ is required"
        sys.exit(1)

rpm_requires = ['python >= 2.6', 'python-daemon', 'pystache', 'python-setuptools']
dependencies = ['python-daemon>=1.5.2,<2.0', 'pystache>=0.4.0', 'setuptools']

_distclass = Distribution
_opts = {
         'build_scripts': { 'executable': '/usr/bin/env python' },
         'bdist_rpm' : { 'requires' : rpm_requires }
        }
_data_files = [('share/doc/%s-%s' % (name, version), ['license/NOTICE.txt', 'license/LICENSE.txt', 'CHANGELOG.txt']),
               ('init/redhat', ['init/redhat/cfn-hup']),
               ('init/ubuntu', ['init/ubuntu/cfn-hup'])]
try:
    import py2exe
    class WindowsDistribution(Distribution):
        def __init__(self, attrs):
            self.com_server = []
            self.ctypes_com_server = []
            self.service = ["cfnbootstrap.winhup"]
            self.isapi = []
            self.windows = []
            self.zipfile = 'library.zip'
            self.console = ['bin/cfn-init', 'bin/cfn-signal', 'bin/cfn-get-metadata', 'bin/cfn-hup', 'bin/cfn-elect-cmd-leader', 'bin/cfn-send-cmd-result', 'bin/cfn-send-cmd-event']
            Distribution.__init__(self, attrs)
    _distclass = WindowsDistribution
    _opts['py2exe'] = {
                        'typelibs' : [('{000C1092-0000-0000-C000-000000000046}', 1033, 1, 0),
                                      ('{E34CB9F1-C7F7-424C-BE29-027DCC09363A}', 0, 1, 0)],
                        'excludes' : ['pyreadline', 'difflib', 'distutils', 'doctest', 'pdb', 'inspect', 'unittest', 'adodbapi'],
                        'includes' : ['dbhash', 'dumbdbm'],
                        'dll_excludes' : ['msvcr71.dll', 'w9xpopen.exe', ''],
                        'compressed' : True,
                      }
    _data_files = [('', ['license/win/NOTICE.txt', 'license/win/LICENSE.rtf', 'cfnbootstrap/packages/requests/cacert.pem', 'cfnbootstrap/resources/documents/endpoints.json', 'CHANGELOG.txt'])]
except ImportError:
    pass


setup(
    distclass = _distclass,
    name=name,
    version=version,
    description='An EC2 bootstrapper for CloudFormation',
    long_description="Bootstraps EC2 instances by retrieving and processing the Metadata block of a CloudFormation resource.",
    author='AWS CloudFormation',
    url='http://aws.amazon.com/cloudformation/',
    license='Apache 2.0',
    classifiers = ['License :: OSI Approved :: Apache Software License'],
    packages=[
        'cfnbootstrap',
        'cfnbootstrap.packages',
        'cfnbootstrap.packages.requests',
        'cfnbootstrap.packages.requests.packages',
        'cfnbootstrap.packages.requests.packages.chardet',
        'cfnbootstrap.packages.requests.packages.urllib3',
        'cfnbootstrap.packages.requests.packages.urllib3.packages',
        'cfnbootstrap.packages.requests.packages.urllib3.contrib',
        'cfnbootstrap.packages.requests.packages.urllib3.util',
        'cfnbootstrap.packages.requests.packages.urllib3.packages.ssl_match_hostname',
        'cfnbootstrap.resources',
        'cfnbootstrap.resources.documents'
    ],
    install_requires=dependencies,
    scripts=['bin/cfn-init', 'bin/cfn-signal', 'bin/cfn-get-metadata', 'bin/cfn-hup', 'bin/cfn-elect-cmd-leader', 'bin/cfn-send-cmd-result', 'bin/cfn-send-cmd-event'],
    data_files=_data_files,
    package_data={'cfnbootstrap': ['packages/requests/cacert.pem', 'resources/documents/endpoints.json']},
    options=_opts
)



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
"""
A library for building an installation from metadata

Classes:
Contractor - orchestrates the build process
Carpenter - does the concrete work of applying metadata to the installation
Tool - performs a specific task on an installation
ToolError - a base exception type for all tools

CloudFormationCarpenter - Orchestrates a non-delegated installation
YumTool - installs packages via yum

"""
from cfnbootstrap import platform_utils
from cfnbootstrap.apt_tool import AptTool
from cfnbootstrap.auth import AuthenticationConfig
from cfnbootstrap.command_tool import CommandTool
from cfnbootstrap.construction_errors import BuildError, NoSuchConfigSetError, \
    NoSuchConfigurationError, CircularConfigSetDependencyError
from cfnbootstrap.file_tool import FileTool
from cfnbootstrap.lang_package_tools import PythonTool, GemTool
from cfnbootstrap.msi_tool import MsiTool
from cfnbootstrap.rpm_tools import RpmTool, YumTool
from cfnbootstrap.service_tools import SysVInitTool, WindowsServiceTool
from cfnbootstrap.sources_tool import SourcesTool
from cfnbootstrap.user_group_tools import GroupTool, UserTool
import cfnbootstrap.json_file_manager as JsonFileManager 
import collections
import contextlib
import logging
import operator
import os.path
import sys
import time

log = logging.getLogger("cfn.init")
cmd_log = logging.getLogger("cfn.init.cmd")

class WorkLog(object):
    """
    Keeps track of pending work, and can resume from the last known point
    Useful for commands that cause restarts
    """

    def __init__(self, dbname='resume_db.json'):
        if os.name == 'nt':
            self._json_db_dir = os.path.expandvars(r'${SystemDrive}\cfn\cfn-init')
        else:
            self._json_db_dir = '/var/lib/cfn-init'

        if not os.path.isdir(self._json_db_dir) and not os.path.exists(self._json_db_dir):
            os.makedirs(self._json_db_dir, 0700)

        if not os.path.isdir(self._json_db_dir):
            print >> sys.stderr, "Could not create %s to store the work log" % self._json_db_dir
            logging.error("Could not create %s to store the work log", self._json_db_dir)

        self._dbname = dbname
        self._jsonConverter = JsonFileManager.Converter([ConfigDefinition])

    def clear(self):
        JsonFileManager.create(self._json_db_dir, self._dbname);

    def clear_except_metadata(self):
        json_data = JsonFileManager.read(self._json_db_dir, self._dbname)
        metadata = json_data.get('metadata', None)
        json_data = {}
        if metadata != None:
            json_data['metadata'] = metadata
        JsonFileManager.write(self._json_db_dir, self._dbname, json_data)

    def put(self, key, data):
        json_data = JsonFileManager.read(self._json_db_dir, self._dbname)
        if data:
            json_data[key] = self._jsonConverter.serialize(data)
        elif key in json_data:
            del json_data[key]
        JsonFileManager.write(self._json_db_dir, self._dbname, json_data)

    def has_key(self, key):
        json_data = JsonFileManager.read(self._json_db_dir, self._dbname)
        return key in json_data

    def get(self, key, default=None):
        json_data = JsonFileManager.read(self._json_db_dir, self._dbname)
        if key in json_data:
            return self._jsonConverter.deserialize(json_data[key])
        else:
            return default

    def delete(self, key):
        json_data = JsonFileManager.read(self._json_db_dir, self._dbname)
        if key in json_data:
            del json_data[key]
            JsonFileManager.write(self._json_db_dir, self._dbname, json_data)

    def pop(self, key):
        json_data = JsonFileManager.read(self._json_db_dir, self._dbname)
        if key in json_data:
            value = self._jsonConverter.deserialize(json_data[key])
            ret_val = value.popleft()
            if not value:
                del json_data[key]
            else:
                json_data[key] = self._jsonConverter.serialize(value)
            JsonFileManager.write(self._json_db_dir, self._dbname, json_data)
            return ret_val
        else:
            return None

    def build(self, metadata, configSets):
        self.put('metadata', metadata)
        platform_utils.set_reboot_trigger()
        Contractor(metadata).build(configSets, self)

    def run_commands(self):
        cmd_tool = CommandTool()
        while self.has_key('commands'):
            next_cmd = self.pop('commands')
            changes = collections.defaultdict(list)
            changes.update(self.get('changes', {}))
            cmd_options = next_cmd[1]
            command_changes = cmd_tool.apply({next_cmd[0]:cmd_options})
            changes['commands'].extend(command_changes)
            self.put('changes', changes)
            if not command_changes:
                log.info("Not waiting as command did not execute")
            else:
                wait = CommandTool.get_wait(cmd_options)
                if wait < 0:
                    log.info("Waiting indefinitely for command to reboot")
                    sys.exit(0)
                elif wait > 0:
                    log.info("Waiting %s seconds for reboot", wait)
                    time.sleep(wait)
            
        for manager, services in self.get('services', {}).iteritems():
            if manager in CloudFormationCarpenter._serviceTools:
                CloudFormationCarpenter._serviceTools[manager]().apply(services, self.get('changes', collections.defaultdict(list)))
            else:
                log.warn("Unsupported service manager: %s", manager)

        if self.has_key('changes'):
            self.delete('changes')

        if self.has_key('services'):
            self.delete('services')

    def resume(self):
        log.debug("Starting resume")
        platform_utils.set_reboot_trigger()

        self.run_commands()

        contractor = Contractor(self.get('metadata'))

        #TODO: apply services when supported by Windows

        while self.has_key('configs'):
            next_config = self.pop('configs')
            log.debug("Resuming config: %s", next_config.name)
            contractor.run_config(next_config, self)

        if self.has_key('configSets'):
            remaining_sets = self.get('configSets')
            log.debug("Resuming configSets: %s", remaining_sets)
            contractor.build(remaining_sets, self)
        else:
            self.clear()
            platform_utils.clear_reboot_trigger()

        log.debug("Resume completed")


class CloudFormationCarpenter(object):
    """
    Takes a model and uses tools to make it reality
    """

    _packageTools = { "yum" : YumTool,
                      "rubygems" : GemTool,
                      "python" : PythonTool,
                      "rpm" : RpmTool,
                      "apt" : AptTool,
                      "msi" : MsiTool }

    _pkgOrder = ["msi", "dpkg", "rpm", "apt", "yum"]

    _serviceTools = { "sysvinit" : SysVInitTool, "windows" : WindowsServiceTool }

    @staticmethod
    def _pkgsort(x, y):
        order = CloudFormationCarpenter._pkgOrder
        if x[0] in order and y[0] in order:
            return cmp(order.index(x[0]), order.index(y[0]))
        elif x[0] in order:
            return -1
        elif y[0] in order:
            return 1
        else:
            return cmp(x[0].lower(), y[0].lower())

    def __init__(self, config, auth_config):
        self._config = config
        self._auth_config = auth_config

    def build(self, worklog):
        changes = collections.defaultdict(list)

        changes['packages'] = collections.defaultdict(list)
        if self._config.packages:
            for manager, packages in sorted(self._config.packages.iteritems(), cmp=CloudFormationCarpenter._pkgsort):
                if manager in CloudFormationCarpenter._packageTools:
                    changes['packages'][manager] = CloudFormationCarpenter._packageTools[manager]().apply(packages, self._auth_config)
                else:
                    log.warn('Unsupported package manager: %s', manager)
        else:
            log.debug("No packages specified")

        if self._config.groups:
            changes['groups'] = GroupTool().apply(self._config.groups)
        else:
            log.debug("No groups specified")

        if self._config.users:
            changes['users'] = UserTool().apply(self._config.users)
        else:
            log.debug("No users specified")

        if self._config.sources:
            changes['sources'] = SourcesTool().apply(self._config.sources, self._auth_config)
        else:
            log.debug("No sources specified")

        if self._config.files:
            changes['files'] = FileTool().apply(self._config.files, self._auth_config)
        else:
            log.debug("No files specified")

        if self._config.commands:
            if os.name=='nt':
                worklog.put('changes', changes)
                worklog.put('commands', collections.deque(sorted(self._config.commands.iteritems(), key=operator.itemgetter(0))))
            else:
                changes['commands'] = CommandTool().apply(self._config.commands)
        else:
            log.debug("No commands specified")

        if self._config.services:
            if os.name=='nt':
                worklog.put('services', self._config.services)
            else:
                for manager, services in self._config.services.iteritems():
                    if manager in CloudFormationCarpenter._serviceTools:
                        CloudFormationCarpenter._serviceTools[manager]().apply(services, changes)
                    else:
                        log.warn("Unsupported service manager: %s", manager)
        else:
            log.debug("No services specified")

class ConfigDefinition(object):
    """
    Encapsulates one config definition
    """

    def __init__(self, name, model):
        self._name = name
        self._files = model.get("files")
        self._packages = model.get("packages")
        self._services = model.get("services")
        self._sources = model.get("sources")
        self._commands = model.get("commands")
        self._users = model.get("users")
        self._groups = model.get("groups")

    @property
    def name(self):
        return self._name

    @property
    def files(self):
        return self._files

    @property
    def packages(self):
        return self._packages

    @property
    def services(self):
        return self._services

    @property
    def sources(self):
        return self._sources

    @property
    def commands(self):
        return self._commands

    @property
    def users(self):
        return self._users

    @property
    def groups(self):
        return self._groups

    def __str__(self):
        return 'Config(%s)' % self._name

    def serialize(self, marker):
        return {marker: self.__dict__}

    @classmethod
    def from_json(cls, json_data):
        model = {}
        for field in json_data:
            prop = field[1:]
            if field == '_name':
                name = json_data[field]
            else:
                model[prop] = json_data[field]
        return cls(name, model)

class ConfigSetRef(object):
    """
    Encapsulates a ref to a ConfigSet
    """

    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name

    def __str__(self):
        return 'ConfigSet(%s)' % self._name

class ConfigSet(object):
    """
    A list of ConfigDefinition or ConfigSetRef objects with their dependencies
    """

    def __init__(self, configDef=None):
        """
        Arguments:
        configDef - optional ConfigDefinition|ConfigSetRef to initialize this list with (handy for 1-member lists)
        """
        self._defs = [] if not configDef else [configDef]
        self._dependencies = set() if (not configDef or isinstance(configDef, ConfigDefinition)) else set([configDef.name])

    def addConfigDef(self, configDef):
        if isinstance(configDef, ConfigSetRef):
            self._dependencies.add(configDef.name)
        self._defs.append(configDef)

    def extend(self, configDefList):
        for cd in configDefList.configDefs:
            self.addConfigDef(cd)

    @property
    def dependencies(self):
        return self._dependencies

    @property
    def configDefs(self):
        return self._defs

    def __str__(self):
        return 'ConfigSet of: %s' % ','.join(self._defs)

class Contractor(object):
    """
    Take in a metadata model and force the environment to match it, returning nothing.

    Processes configSets if they exist; otherwise, invents a virtual configSet named
    "default" with one config of "config"

    """

    _configKey = "AWS::CloudFormation::Init"
    _authKey = "AWS::CloudFormation::Authentication"
    _configSetsKey = "configSets"

    def __init__(self, model):
        initModel = model.get(Contractor._configKey)
        if not initModel:
            raise ValueError("Metadata does not contain '%s'" % Contractor._configKey)

        if not Contractor._configSetsKey in initModel:
            self._configSets = { 'default' : [ConfigDefinition("config", initModel.get("config", dict()))]}
        else:
            configSetsDef = initModel[Contractor._configSetsKey]
            if not isinstance(configSetsDef, dict):
                raise ValueError("%s should be a mapping of name to list" % Contractor._configSetsKey)

            self._processConfigSetsDefinition(configSetsDef, initModel)

        self._auth_config = AuthenticationConfig(model.get(Contractor._authKey, {}))

    def _processConfigSetsDefinition(self, configSetsDef, model):
        """
        Parse a set of configSets from the model and collapse them, validating there are no cycles
        and that all references are valid.
        """

        # This builds both a map of the uncollapsed config sets
        # as well as a lookup and reverse lookup table
        # so we can traverse the graph and detect cycles
        # in a not-terrible time

        rawConfigSets = {}
        dependencyTree = {} # maps configSets to the configSets they depend on
        reverseDependencyTree = collections.defaultdict(set) # maps configSets to the configSets that depend on them
        roots = set() # the roots of the configSets graph -- configSets without dependencies
        for configSetName, configList in configSetsDef.iteritems():
            processedList = self._processConfigList(configList, model)
            if processedList.dependencies:
                dependencyTree[configSetName] = set(processedList.dependencies)
                for dependency in processedList.dependencies:
                    reverseDependencyTree[dependency].add(configSetName)
            else:
                roots.add(configSetName)

            rawConfigSets[configSetName] = list(processedList.configDefs)

        if not roots:
            raise CircularConfigSetDependencyError("No configSets exist without references; this creates a circular dependency and is not allowed")

        self._configSets = {}
        # use a traditional (Kahn) topological sort to traverse the configSets in dependency order
        # http://en.wikipedia.org/wiki/Topological_sort#Algorithms has a nice description
        while roots:
            configSet = roots.pop()
            self._configSets[configSet] = self._collapse(configSet, rawConfigSets[configSet])
            for dependent in reverseDependencyTree.pop(configSet, []):
                dependencyTree[dependent].remove(configSet)
                if not dependencyTree[dependent]:
                    roots.add(dependent)
                    del dependencyTree[dependent]

        if dependencyTree:
            raise CircularConfigSetDependencyError("At least one circular dependency detected; this is not allowed. Culprits: " + ', '.join(dependencyTree.keys()))


    def _collapse(self, configSetName, configList):
        """
        Transform ConfigSetRefs into the contents of the ConfigSets they reference, returning a list of only ConfigDefinition objects
        """
        returnList = []

        for config in configList:
            if isinstance(config, ConfigDefinition):
                returnList.append(config)
            else:
                if not config.name in self._configSets:
                    raise ValueError("ConfigSet %s referenced ConfigSet %s but it is not defined" % (configSetName, config.name))
                returnList.extend(self._configSets[config.name])

        return returnList

    def _processConfigList(self, configList, model):
        """
        Processes a parsed-JSON list of config definitions, returning a ConfigSet

        Handles both references ({"ConfigSet" : "name"}) and plain config names
        so users can define simple ConfigSets without using lists, and so we can recurse simply
        """

        if isinstance(configList, basestring):
            if not configList in model:
                raise NoSuchConfigurationError("No configuration found with name: %s" % configList)
            return ConfigSet(ConfigDefinition(configList, model[configList]))

        if isinstance(configList, dict):
            if not 'ConfigSet' in configList:
                raise ValueError("Config definitions must be either a config name or a reference in the format {'ConfigSet':<config set name>}")
            setName = configList['ConfigSet']
            if not setName in model[Contractor._configSetsKey]:
                raise ValueError("Configuration set %s was referenced but not defined" % setName)
            return ConfigSet(ConfigSetRef(setName))

        returnSet = ConfigSet()
        for configDef in configList:
            returnSet.extend(self._processConfigList(configDef, model))

        return returnSet


    def build(self, configSets, worklog):
        """Does the work described by each configSet, in order, returning nothing"""

        worklog.clear_except_metadata()

        configSets = collections.deque(configSets)
        log.info("Running configSets: %s", ', '.join(configSets))

        while configSets:
            configSetName = configSets.popleft()
            if not configSetName in self._configSets:
                raise NoSuchConfigSetError("Error: no ConfigSet named %s exists" % configSetName)

            worklog.put('configSets', configSets)

            configSet = collections.deque(self._configSets[configSetName])
            log.info("Running configSet %s", configSetName)
            cmd_log.info("*"*60)
            cmd_log.info("ConfigSet %s", configSetName)
            while configSet:
                config = configSet.popleft()

                worklog.put('configs', configSet)

                self.run_config(config, worklog)

        log.info("ConfigSets completed")
        worklog.clear()
        platform_utils.clear_reboot_trigger()

    def run_config(self, config, worklog):
        log.info("Running config %s", config.name)
        cmd_log.info("+"*60)
        cmd_log.info("Config %s", config.name)
        
        try:
            CloudFormationCarpenter(config, self._auth_config).build(worklog)

            worklog.run_commands()
        except BuildError, e:
            log.exception("Error encountered during build of %s: %s", config.name, str(e))
            raise

    @classmethod
    def metadataValid(cls, metadata):
        return metadata and cls._configKey in metadata and metadata[cls._configKey]

    @property
    def configs(self):
        return dict(self._configSets)



#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
from cfnbootstrap import util
from cfnbootstrap.construction_errors import ToolError
from cfnbootstrap.util import ProcessHelper
import collections
import logging
import os.path

log = logging.getLogger("cfn.init")

_windows_supported = True
try:
    import win32service
    import win32serviceutil
except ImportError:
    _windows_supported = False

class ServiceTool(object):

    def _detect_required_restart(self, serviceProperties, changes):
        if self._list_type_change_occurred(serviceProperties, changes, 'files'):
            return True

        if self._list_type_change_occurred(serviceProperties, changes, 'sources'):
            return True

        if self._list_type_change_occurred(serviceProperties, changes, 'groups'):
            return True

        if self._list_type_change_occurred(serviceProperties, changes, 'users'):
            return True

        if self._list_type_change_occurred(serviceProperties, changes, 'commands'):
            return True

        if 'packages' in serviceProperties and 'packages' in changes:
            for manager, pkg_list in changes['packages'].iteritems():
                if manager in serviceProperties['packages']:
                    if frozenset(serviceProperties['packages'][manager]) & frozenset(pkg_list):
                        return True

        return False

    def _list_type_change_occurred(self, serviceProperties, changes, key):
        if key in serviceProperties and key in changes:
            if frozenset(serviceProperties[key]) & frozenset(changes[key]):
                return True

        return False

class WindowsServiceTool(ServiceTool):
    """
    Manages Windows services
    
    """
    
    def apply(self, action, changes = collections.defaultdict(list)):
        """
        Takes a dict of service name to dict.
        Keys we look for are:
            - "enabled" (setting a service to "Automatic")
            - "ensureRunning" (actually start the service)
        """

        if not action.keys():
            log.debug("No Windows services specified")
            return
        
        if not _windows_supported:
            raise ToolError("Cannot modify windows services without pywin32")
        
        manager = win32service.OpenSCManager(None, None, win32service.SC_MANAGER_ALL_ACCESS)
        try:
            for service, serviceProperties in action.iteritems():
                handle = win32service.OpenService(manager, service, win32service.SERVICE_ALL_ACCESS)
                try:
                    if "enabled" in serviceProperties:
                        start_type = win32service.SERVICE_AUTO_START if util.interpret_boolean(serviceProperties["enabled"]) else win32service.SERVICE_DEMAND_START
                        self._set_service_startup_type(handle, start_type)
                    else:
                        log.debug("Not modifying enabled state of service %s", service)
                        
                    if self._detect_required_restart(serviceProperties, changes):
                        log.debug("Restarting %s due to change detected in dependency", service)
                        win32serviceutil.RestartService(service)
                    elif "ensureRunning" in serviceProperties:
                        ensureRunning = util.interpret_boolean(serviceProperties["ensureRunning"])
                        status = win32service.QueryServiceStatus(handle)[1]
                        isRunning = status & win32service.SERVICE_RUNNING or status & win32service.SERVICE_START_PENDING
                                        
                        if ensureRunning and not isRunning:
                            log.debug("Starting service %s as it is not running", service)
                            win32service.StartService(handle, None)
                        elif not ensureRunning and isRunning:
                            log.debug("Stopping service %s as it is running", service)
                            win32service.ControlService(handle, win32service.SERVICE_CONTROL_STOP)
                        else:
                            log.debug("No need to modify running state of service %s", service)
                    else:
                        log.debug("Not modifying running state of service %s", service)
                finally:
                    win32service.CloseServiceHandle(handle)
        finally:
            win32service.CloseServiceHandle(manager)
    
    def _set_service_startup_type(self, service, start_type):
        win32service.ChangeServiceConfig(service,
                                         win32service.SERVICE_NO_CHANGE,
                                         start_type,
                                         win32service.SERVICE_NO_CHANGE,
                                         None,
                                         None,
                                         0,
                                         None,
                                         None,
                                         None,
                                         None)

class SysVInitTool(ServiceTool):
    """
    Manages SysV Init services

    """

    def apply(self, action, changes = collections.defaultdict(list)):
        """
        Takes a dict of service name to dict.
        Keys we look for are:
            - "enabled" (equivalent to /sbin/chkconfig service on)
            - "ensureRunning" (equivalent to /sbin/service service start)
        """

        if not action.keys():
            log.debug("No System V init scripts specified")
            return

        for service, serviceProperties in action.iteritems():
            force_restart = self._detect_required_restart(serviceProperties, changes)

            if "enabled" in serviceProperties:
                self._set_service_enabled(service, util.interpret_boolean(serviceProperties["enabled"]))
            else:
                log.debug("Not modifying enabled state of service %s", service)

            if force_restart:
                log.debug("Restarting %s due to change detected in dependency", service)
                self._restart_service(service)
            elif "ensureRunning" in serviceProperties:
                ensureRunning = util.interpret_boolean(serviceProperties["ensureRunning"])
                isRunning = self._is_service_running(service)
                if ensureRunning and not isRunning:
                    log.debug("Starting service %s as it is not running", service)
                    self._start_service(service)
                elif not ensureRunning and isRunning:
                    log.debug("Stopping service %s as it is running", service)
                    self._stop_service(service)
                else:
                    log.debug("No need to modify running state of service %s", service)
            else:
                log.debug("Not modifying running state of service %s", service)

    def _restart_service(self, service):
        cmd = self._get_service_executable(service)
        cmd.append("restart")

        result = ProcessHelper(cmd).call()

        if result.returncode:
            log.error("Could not restart service %s; return code was %s", service, result.returncode)
            log.debug("Service output: %s", result.stdout)
            raise ToolError("Could not restart %s" % service)
        else:
            log.info("Restarted %s successfully", service)

    def _start_service(self, service):
        cmd = self._get_service_executable(service)
        cmd.append("start")

        result = ProcessHelper(cmd).call()

        if result.returncode:
            log.error("Could not start service %s; return code was %s", service, result.returncode)
            log.debug("Service output: %s", result.stdout)
            raise ToolError("Could not start %s" % service)
        else:
            log.info("Started %s successfully", service)

    def _stop_service(self, service):
        cmd = self._get_service_executable(service)
        cmd.append("stop")

        result = ProcessHelper(cmd).call()

        if result.returncode:
            log.error("Could not stop service %s; return code was %s", service, result.returncode)
            log.debug("Service output: %s", result.stdout)
            raise ToolError("Could not stop %s" % service)
        else:
            log.info("Stopped %s successfully", service)

    def _is_service_running(self, service):
        cmd = self._get_service_executable(service)
        cmd.append("status")

        result = ProcessHelper(cmd).call()

        if result.returncode:
            return False

        return True

    def _set_service_enabled(self, service, enabled=True):
        modifier = self._get_service_modifier()
        if not modifier:
            log.error("Could not enable %s, as chkconfig and update-rc.d were not available", service)
            return

        log.debug("Setting service %s to %s", service, "enabled" if enabled else "disabled")

        modifier.set_service_enabled(service, enabled);

    def _get_service_modifier(self):
        if hasattr(self, '_cached_modifier'):
            return self._cached_modifier

        if Chkconfig.installed():
            self._cached_modifier = Chkconfig()
        elif UpdateRcD.installed():
            self._cached_modifier = UpdateRcD()
        else:
            self._cached_modifier = None

        log.debug("Using service modifier: %s", self._cached_modifier)

        return self._cached_modifier

    def _get_service_executable(self, service):
        """
        Returns the service executable as an array
        Right now either ["/sbin/service", service], ["/usr/sbin/service", service], or ["/etc/init.d/<service>"]
        """
        if not hasattr(self, '_service_runner'):
            if os.path.exists("/sbin/service"):
                self._service_runner = "/sbin/service"
            elif os.path.exists("/usr/sbin/service"):
                self._service_runner = "/usr/sbin/service"
            else:
                self._service_runner = None

            if self._service_runner:
                log.debug("Using service runner: %s", self._service_runner)
            else:
                log.debug("Running init scripts directly")

        if self._service_runner:
            return [self._service_runner, service]
        else:
            return ["/etc/init.d/%s" % service]

class Chkconfig(object):

    _executable = "/sbin/chkconfig"

    @classmethod
    def installed(cls):
        return os.path.exists(cls._executable)

    def __init__(self):
        #Leaving this open for multiple locations of chkconfig
        self._executable = Chkconfig._executable

    def __str__(self, *args, **kwargs):
        return self._executable

    def set_service_enabled(self, service, enabled=True):
        if not os.path.exists(self._executable):
            raise ToolError("Cannot find chkconfig")

        result = ProcessHelper([self._executable, service, 'on' if enabled else 'off']).call()

        if result.returncode:
            log.error("chkconfig failed with error %s. Output: %s", result.returncode, result.stdout)
            raise ToolError("Could not %s service %s" % ("enable" if enabled else "disable", service), result.returncode)
        else:
            log.info("%s service %s", "enabled" if enabled else "disabled", service)

class UpdateRcD(object):

    _executable = "/usr/sbin/update-rc.d"

    @classmethod
    def installed(cls):
        return os.path.exists(cls._executable)

    def __init__(self):
        #Leaving this open for multiple locations of update-rc.d
        self._executable = UpdateRcD._executable

    def __str__(self, *args, **kwargs):
        return self._executable

    def set_service_enabled(self, service, enabled=True):
        if not os.path.exists(self._executable):
            raise ToolError("Cannot find update-rc.d")

        result = ProcessHelper([self._executable, service, 'enable' if enabled else 'disable']).call()

        if result.returncode:
            log.error("update-rc.d failed with error %s. Output: %s", result.returncode, result.stdout)
            raise ToolError("Could not %s service %s" % ("enable" if enabled else "disable", service), result.returncode)
        else:
            log.info("%s service %s", "enabled" if enabled else "disabled", service)




#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
import endpoint_tool
import urllib
import operator
from cfnbootstrap.resources import documents
from cfnbootstrap.packages.requests.auth import AuthBase, HTTPBasicAuth
import base64
import datetime
import hashlib
import hmac
import logging
import re
import urlparse
import util

_AMZ_CONTENT_SHA_HEADER = 'x-amz-content-sha256'
_AMZ_DATE_HEADER = 'x-amz-date'
_AUTHORIZATION_HEADER = 'Authorization'
_SECURITY_TOKEN_HEADER = 'x-amz-security-token'

log = logging.getLogger("cfn.init")

def _extract_bucket_from_url(unparsed_url):
    endpoint = endpoint_tool.get_endpoint_for_url("AmazonS3", unparsed_url)
    if endpoint is None:
        # Not an S3 URL, skip
        return None

    bucket = endpoint.get_subdomain_prefix(unparsed_url)
    if not bucket:
        # This means that we're using path-style buckets
        # lop off the first / and return everything up to the next /
        return urlparse.urlparse(unparsed_url).path[1:].partition('/')[0]
    else:
        # Subdomain-style S3 URL
        # Remove the trailing dot if it exists
        return bucket.rstrip('.')

class S3Signer(object):

    def __init__(self, creds, region=None):
        self._creds = creds
        self._region = region
        self._nowfunction = datetime.datetime.utcnow

    # Inspired by digest auth handling: https://github.com/kennethreitz/requests/blob/v2.6.0/requests/auth.py#L172
    def handle_redirect(self, r, **kwargs):
        if r.is_redirect and 'Location' in r.headers:
            if _extract_bucket_from_url(r.headers['Location']) is not None:
                log.warn("Handling redirect to S3 location: %s", r.headers['Location'])

                # Consume content and release the original connection
                # to allow our new request to reuse the same one.
                r.content
                r.raw.release_conn()
                prep = r.request.copy()

                # Remove V2 S3 auth headers
                prep.headers.pop(_SECURITY_TOKEN_HEADER, '')
                prep.headers.pop(_AUTHORIZATION_HEADER, '')
                prep.headers.pop(_AMZ_DATE_HEADER, '')
                prep.headers.pop(_AMZ_CONTENT_SHA_HEADER, '')

                # Accept the S3 redirect
                prep.url = r.headers['Location']

                prep = self.sign(prep)

                _r = r.connection.send(prep, **kwargs)
                _r.history.append(r)
                _r.request = prep
                return _r
            else:
                log.error('S3 redirected to non-S3 url: %s', r.headers['Location'])

        return r

    def sign(self, req):
        # S3 will redirect requests to the appropriate endpoint for a bucket.  This response handler will handle that redirect
        req.register_hook('response', self.handle_redirect)

        # Requests only quotes illegal characters in a URL, leaving reserved chars.
        # We want to fully quote the URL, so we first unquote the url before requoting it in our signature calculation
        full_url = urllib.unquote(req.full_url if hasattr(req, 'full_url') else req.url)

        region = self._region
        if not region:
            endpoint = endpoint_tool.get_endpoint_for_url("AmazonS3", full_url)
            if endpoint.is_default:
                log.warn('Falling back to Signature Version 2 as no region was specified in S3 URL')
                return S3V2Signer(self._creds).sign(req)
            region = endpoint.region

        parsed = urlparse.urlparse(full_url)

        timestamp = self._nowfunction()
        timestamp_formatted = timestamp.strftime('%Y%m%dT%H%M%SZ')
        timestamp_short = timestamp.strftime('%Y%m%d')

        scope =  timestamp_short + '/' + region + '/s3/aws4_request'

        req.headers[_AMZ_DATE_HEADER] = timestamp_formatted
        if self._creds.security_token:
            req.headers[_SECURITY_TOKEN_HEADER] = self._creds.security_token
        req.headers['host'] = parsed.netloc

        hashed_payload = hashlib.sha256(req.body if req.body is not None else '').hexdigest()
        req.headers[_AMZ_CONTENT_SHA_HEADER] = hashed_payload

        canonical_request = req.method + '\n'
        canonical_request += self._canonicalize_uri(full_url) + '\n'
        canonical_request += self._canonicalize_query(urlparse.parse_qs(parsed.query, True)) + '\n'

        headers_to_sign = req.headers.copy()
        (canonical_headers, signed_headers) = self._canonicalize_headers(headers_to_sign)
        canonical_request += canonical_headers + '\n' + signed_headers + '\n'
        canonical_request += hashed_payload

        string_to_sign = 'AWS4-HMAC-SHA256\n' + timestamp_formatted + '\n' + scope + '\n' + hashlib.sha256(canonical_request.encode('utf-8')).hexdigest()

        derived_key = hmac.new(("AWS4" + self._creds.secret_key).encode('utf-8'), timestamp_short.encode('utf-8'), hashlib.sha256).digest()
        derived_key = hmac.new(derived_key, region.encode('utf-8'), hashlib.sha256).digest()
        derived_key = hmac.new(derived_key, 's3'.encode('utf-8'), hashlib.sha256).digest()
        derived_key = hmac.new(derived_key, "aws4_request".encode('utf-8'), hashlib.sha256).digest()

        signature = hmac.new(derived_key, string_to_sign.encode('utf-8'), hashlib.sha256).hexdigest()

        credential = self._creds.access_key + '/' + scope
        req.headers[_AUTHORIZATION_HEADER] = 'AWS4-HMAC-SHA256 Credential=%s,SignedHeaders=%s,Signature=%s' % (credential, signed_headers, signature)

        return req

    def _canonicalize_uri(self, uri):
        split = urlparse.urlsplit(uri)
        if not split.path:
            return '/'
        path = urlparse.urlsplit(urlparse.urljoin('http://foo.com', split.path.lstrip('/'))).path
        return urllib.quote(path, '/~') if path else '/'

    def _canonicalize_query(self, params):
        if not params:
            return ''
        encoded_pairs = []
        for entry in params.iteritems():
            for value in entry[1]:
                encoded_pairs.append((urllib.quote(entry[0], '~'), urllib.quote(value, '~') if value else ''))

        sorted_pairs = sorted(encoded_pairs, key=operator.itemgetter(0, 1))

        return '&'.join(('='.join(pair) for pair in sorted_pairs))

    def _canonicalize_headers(self, headers):
        canon_headers = {}
        for key, value in ((key.lower(), re.sub(r'(?su)[\s]+', ' ', value).strip()) for key, value in headers.iteritems()):
            if key in canon_headers:
                canon_headers[key] = canon_headers[key] + ',' + value
            else:
                canon_headers[key] = value

        sorted_entries = sorted(canon_headers.iteritems(), key=operator.itemgetter(0))

        return '\n'.join((':'.join(entry) for entry in sorted_entries)) + '\n', ';'.join((entry[0] for entry in sorted_entries))


class S3V2Signer(object):

    def __init__(self, creds):
        self._creds = creds

    def sign(self, req):
        if 'Date' not in req.headers:
            req.headers[_AMZ_DATE_HEADER] = datetime.datetime.utcnow().replace(microsecond=0).strftime("%a, %d %b %Y %H:%M:%S GMT")

        if self._creds.security_token:
            req.headers[_SECURITY_TOKEN_HEADER] = self._creds.security_token

        stringToSign = req.method + '\n'
        stringToSign += req.headers.get('content-md5', '') + '\n'
        stringToSign += req.headers.get('content-type', '') + '\n'
        stringToSign += req.headers.get('date', '') + '\n'
        stringToSign += self._canonicalize_headers(req)
        stringToSign += self._canonicalize_resource(req)

        signed = base64.encodestring(hmac.new(self._creds.secret_key.encode('utf-8'), stringToSign.encode('utf-8'), hashlib.sha1).digest()).strip()

        req.headers[_AUTHORIZATION_HEADER] = 'AWS %s:%s' % (self._creds.access_key, signed)

        return req

    def _canonicalize_headers(self, req):
        headers = [(hdr.lower(), val) for hdr, val in req.headers.iteritems() if hdr.lower().startswith('x-amz')]
        return '\n'.join([hdr + ':' + val for hdr, val in sorted(headers)]) + '\n' if headers else ''

    def _canonicalize_resource(self, req):
        unparsed_url = req.full_url if hasattr(req, 'full_url') else req.url
        endpoint = endpoint_tool.get_endpoint_for_url("AmazonS3", unparsed_url)
        bucket = endpoint.get_subdomain_prefix(unparsed_url)
        url = urlparse.urlparse(unparsed_url)
        if not bucket:
            # Path-style
            return url.path
        else:
            # Subdomain-style
            return '/' + bucket.rstrip('.') + url.path

class S3DefaultAuth(AuthBase):

    def __init__(self):
        self._bucketToAuth = {}

    def add_auth_for_bucket(self, bucket, auth_impl):
        self._bucketToAuth[bucket] = auth_impl

    def __call__(self, req):
        bucket = self._extract_bucket(req)
        if bucket and bucket in self._bucketToAuth:
            return self._bucketToAuth[bucket](req)
        return req

    def _extract_bucket(self, req):
        return _extract_bucket_from_url(req.full_url if hasattr(req, 'full_url') else req.url)


class S3RoleAuth(AuthBase):

    def __init__(self, roleName):
        self._roleName=roleName

    def __call__(self, req):
        return S3Signer(util.get_role_creds(self._roleName)).sign(req)

class S3Auth(AuthBase):

    def __init__(self, access_key, secret_key):
        self._signer = S3Signer(util.Credentials(access_key, secret_key))

    def __call__(self, req):
        return self._signer.sign(req)

class BasicDefaultAuth(AuthBase):

    def __init__(self):
        self._auths = {}

    def __call__(self, req):
        base_uri = urlparse.urlparse(req.full_url if hasattr(req, 'full_url') else req.url).netloc
        if base_uri in self._auths:
            return self._auths[base_uri](req)
        return req

    def add_password(self, uri, username, password):
        self._auths[uri] = HTTPBasicAuth(username, password)

class DefaultAuth(AuthBase):

    def __init__(self, s3, basic):
        self._s3 = s3
        self._basic = basic

    def __call__(self, req):
        return self._s3(self._basic(req))

class AuthenticationConfig(object):

    def __init__(self, model):

        self._auths = {}

        s3Auth = S3DefaultAuth()
        basicAuth = BasicDefaultAuth()

        for key, config in model.iteritems():
            configType = config.get('type', '')
            if 's3' == configType.lower():
                auth_impl = None
                if 'accessKeyId' in config and 'secretKey' in config:
                    auth_impl = S3Auth(config['accessKeyId'], config['secretKey'])
                elif 'roleName' in config:
                    auth_impl = S3RoleAuth(config['roleName'])
                else:
                    log.warn('S3 auth requires either "accessKeyId" and "secretKey" or "roleName"')
                    continue

                self._auths[key] = auth_impl

                if 'buckets' in config:
                    buckets = [config['buckets']] if isinstance(config['buckets'], basestring) else config['buckets']
                    for bucket in buckets:
                        s3Auth.add_auth_for_bucket(bucket, auth_impl)

            elif 'basic' == configType.lower():
                self._auths[key] = HTTPBasicAuth(config.get('username'), config.get('password'))
                if 'uris' in config:
                    if isinstance(config['uris'], basestring):
                        basicAuth.add_password(config['uris'], config.get('username'), config.get('password'))
                    else:
                        for u in config['uris']:
                            basicAuth.add_password(u, config.get('username'), config.get('password'))
            else:
                log.warn("Unrecognized authentication type: %s", configType)

        self._defaultAuth = DefaultAuth(s3Auth, basicAuth)

    def get_auth(self, key):
        if not key or not key in self._auths:
            return self._defaultAuth

        return self._auths[key]
